/*
 * This file is part of the QPxTool project.
 * Copyright (C) 2008-2009 Gennady "ShultZ" Kozlov <qpxtool@mail.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 *
 */

#include <QApplication>
#include <QTreeWidgetItem>
#include <QProcess>
#include <QTcpSocket>
#include <QDataStream>
#include <QXmlStreamReader>
#include <QXmlStreamWriter>

#include <qpxiodevice.h>
#include "device.h"

#include <sys/types.h>
#include <signal.h>

#include <qpx_mmc_defs.h>

#ifndef QT_NO_DEBUG
#include <QDebug>
static int devcnt=0;
#endif

#if defined(_WIN32)
#include <windows.h>
//#if !defined(__MINGW32__)
//#endif
#endif

#define RECALC_ON_LOAD 1

void ErrcADD(Errc<int64_t> *tot, const Errc<int>& o) {
	for (int i=0; i<8; i++)
		if (o.raw.err[i] > 0) tot->raw.err[i] += o.raw.err[i];
};

void ErrcMAX(Errc<int> *max, const Errc<int>& o) {
	for (int i=0; i<8; i++)
		if (max->raw.err[i] < o.raw.err[i]) max->raw.err[i] = o.raw.err[i];
};

void CDErrcAVG(Errc<float> *avg, Errc<int64_t> *tot, uint64_t blocks) {
	if (!blocks) blocks=1;
	for (int i=0; i<8; i++)
		avg->raw.err[i] = (float) tot->raw.err[i] / blocks;
};

void DVDErrcAVG(Errc<float> *avg, Errc<int64_t> *tot, uint64_t blocks) {
	avg->dvd.pie = (float) tot->dvd.pie / blocks;
	avg->dvd.pif = (float) tot->dvd.pif / blocks;
	avg->dvd.poe = (float) tot->dvd.poe / blocks;
	avg->dvd.pof = (float) tot->dvd.pof / blocks;
	avg->dvd.uncr = (float) tot->dvd.uncr / blocks;
	if (blocks >= 8) {
		avg->dvd.pi8 = (float) tot->dvd.pi8 / ( blocks >> 3);
		avg->dvd.po8 = (float) tot->dvd.po8 / ( blocks >> 3);
	}
};

static device NullDev(NULL);

/*
 *
 * Device List
 *
 */

devlist::devlist()
	: QList<device*>()
{
#ifndef QT_NO_DEBUG
	qDebug("* STA: devlist()");
#endif
	devidx=-1;
#ifndef QT_NO_DEBUG
	qDebug("* END: devlist()");
#endif
};

devlist::~devlist() {
#ifndef QT_NO_DEBUG
	qDebug("* STA: ~devlist()");
#endif
	clear();
#ifndef QT_NO_DEBUG
	qDebug("* END: ~devlist()");
#endif
};

int devlist::idx() { return devidx; };

void devlist::setIdx(int iidx) {
	if (iidx<0 || iidx>=size()) {
		devidx=-1;
		return; 
	}
	devidx=iidx;
};

void devlist::clear() {
		while (size()) delete takeLast();
		devidx=-1; 
};

device* devlist::current() {
	if (devidx<0 || devidx>=size()) return &NullDev;
	return (*this)[devidx];
};

/*
 *
 * Device
 *
 */

device::device(QObject* p)
	: QObject(p)
{
#ifndef QT_NO_DEBUG
	qDebug() << "* STA: device(): " << this << " #"<< devcnt++ << " parent: "<< p;
#endif
	preservePluginsList=0;
	running=0;
	type = DevtypeNone;
	host = "";
	port = 0;
	info_set = 0;

	rpc_phase = -1;
	rpc_reg   = -1;
	rpc_ch    = -1;
	rpc_rst   = -1;

	plextor_lock = 0;

	asdb.clear();
	clearMinfo();

	cap	   = 0;
	cap_rd = 0;
	cap_wr = 0;

	life_dn = -1;
	features.supported = 0;
	features.enabled = 0;
	features.grec = 1.0;
	features.vrec_cd_pwr = 0;
	features.vrec_cd_str = 0;
	features.vrec_dvd_pwr = 0;
	features.vrec_dvd_str = 0;
	features.tattoo_inner = 22;
	features.tattoo_outer = 54;

	features.as_mode = AS_MODE_AUTO;
	features.pioq_quiet = PIOQ_STD;

	test_cap=0;
	test_req=0;
	tests = 0;
	ctest = 0;
	WT_simul = 1;

	tspeeds.rt = 1;
	tspeeds.wt = 1;
	tspeeds.errc = 1;
	tspeeds.jb = 1;
	tspeeds.ft = 1;

	io = new QPxIODevice(this);
	proc = NULL;
	sock = NULL;
	mutex = new QMutex();
	io_mutex = new QMutex();

#ifndef QT_NO_DEBUG
	qDebug("* END: device()");
#endif
};

device::~device()
{
#ifndef QT_NO_DEBUG
	qDebug() << "* STA: ~device(): " << this << " #"<< --devcnt;
#endif
	QTreeWidgetItem* item;
#ifdef MINFO_TREE
	while (info_media.size()) {
		item = info_media.takeLast();
		if (item) delete item;
	}
#endif
//	mutex->unlock();
	delete mutex;
#ifndef QT_NO_DEBUG
	qDebug() << "* END: ~device()";
#endif
};

bool device::isRunning()
{
	return running;
};

void device::clearMinfo()
{
	QTreeWidgetItem* item;

	testData.clear();

//	media.isCD      = 0;
//	media.isDVD     = 0;
	media.type		= "-";
	media.category	= "-";
	media.mid		= "-";
	media.erasable	= "-";
	media.layers	= "-";
	media.ilayers	= 1;
	media.prot		= "-";
	media.regions	= "-";
	media.creads	= 0;
	media.creadm	= 0;
	media.creadmsf	= "";
	media.cfrees	= 0;
	media.cfreem	= 0;
	media.cfreemsf	= "";
	media.ctots		= 0;
	media.ctotm		= 0;
	media.ctotmsf	= "";
	media.dstate	= "-";
	media.sstate	= "-";
	media.writer	= "-";
	media.grec      = 0.0;
	media.rspeeds.clear();
	media.wspeedsd.clear();
	media.wspeedsm.clear();

	media.tspeeds_errc.clear();
	media.tspeeds_jb.clear();
#ifdef MINFO_TREE
	while (info_media.size()) {
		item = info_media.takeLast();
		if (item) delete item;
	}
#endif
};

bool device::start()
{
#ifndef QT_NO_DEBUG
	qDebug() << "STA: device::start(" << threadType << ")";
#endif
	if (!mutex->tryLock()) {
#ifndef QT_NO_DEBUG
		qDebug() << "Device busy: " << id;
		qDebug() << "END: device::start()";
#endif
		return false;
	}
	stop=0;
	running=1;
	switch (threadType) {
		case threadDevice:
		case threadMedia:
		case threadGetFeatures:
		case threadGetASDB:
		case threadMQCK:
		case threadAScre:
		case threadDestruct:
		case threadTattoo:
			start_update_info();
			break;
		case threadTest:
			next_test();
			break;
		default:
			mutex->unlock();
#ifndef QT_NO_DEBUG
			qDebug() << "END: device::start()";
#endif
			running=0;
			return false;
	}

#ifndef QT_NO_DEBUG
	qDebug() << "END: device::start()";
#endif
	return true;
}

bool device::start_update_info()
{
#ifndef QT_NO_DEBUG
	qDebug() << "STA: device::start_update_info(" << threadType << ")";
#endif	
	if (threadType==threadMedia) {
		clearMinfo();
//		mwidget->update();
#warning !!! mwidget->clearMedia() call
		//mwidget->clearMedia();
	}

	//if (devices.idx()<0 || devices.idx()>= devices.size() || (threadType!=infoDevice && threadType!=infoMedia)) {
	if (type == DevtypeNone) {
		goto update_info_err;
	}
//	if (threadType == infoDevice) {
//		
//	}
	if (proc) {
		delete proc;
		proc = NULL;
	}
	if (sock) {
		delete sock;
		sock = NULL;
	}

	QObject::connect(io, SIGNAL(readyReadLine()),
		this, SLOT(qscan_process_info()));

	if (type == DevtypeLocal) {
#ifndef QT_NO_DEBUG
		qDebug() << "device: LOCAL";
#endif
		proc = new QProcess(this);
#ifndef QT_NO_DEBUG
		qDebug() << "process created";
#endif
		io->setIODevice(proc);
		proc->setReadChannel(QProcess::StandardOutput);

//		QObject::connect(proc, SIGNAL(readyReadStandardOutput()),
//				this, SLOT(qscan_process_info()));

		switch (threadType) {
			case threadDevice:
				proc->start("qscan", QStringList() << "-d" << path << "-Ip");
				break;
			case threadMedia:
				if (plugin.isEmpty()) {
					proc->start("qscan", QStringList() << "-d" << path << "-m");
				} else {
					proc->start("qscan", QStringList() << "-d" << path << "--force-plugin" << plugin << "-m");
				}
				break;
			case threadGetFeatures:
				proc->start("cdvdcontrol", QStringList() << "-d" << path << "-c");
				break;
			case threadGetASDB:
				proc->start("cdvdcontrol", QStringList() << "-d" << path << "--as-list");
				break;
			case threadMQCK:
				{
					QStringList cdvdopts;
					cdvdopts << "-d" << path;
					cdvdopts << "--mqck" << ((features.as_act_mode & ASMQCK_ADV) ? "advanced" : "quick");
					cdvdopts << "--mqck-speed" << QString::number(features.as_mqckspd);

					qDebug() << cdvdopts;
					proc->start("cdvdcontrol", cdvdopts );
				}
				break;
			case threadAScre:
				{
					QStringList cdvdopts;
					cdvdopts << "-d" << path << "--as-create";
					cdvdopts << ((features.as_act_mode & ASCRE_FULL) ? "f" : "q");
					cdvdopts << ((features.as_act_mode & ASCRE_REPLACE) ? "r" : "a");

					qDebug() << cdvdopts;
					proc->start("cdvdcontrol", cdvdopts );
				}
				break;
			case threadDestruct:
				proc->start("cdvdcontrol", QStringList() << "-d" << path << "--destruct" << (features.as_act_mode ? "full" : "quick"));
				break;
			case threadTattoo:
				proc->start("f1tattoo", QStringList() << "-d" << path << "--tattoo-raw" << features.tattoo_file);
				break;
			default:
				goto update_info_err;
		}
		if (!proc->waitForStarted(10000)) {
#ifndef QT_NO_DEBUG
			qDebug("Can't start qscan!");
#endif
			goto update_info_err;
		}
		QObject::connect(proc, SIGNAL(finished(int, QProcess::ExitStatus)),
				this, SLOT(qscan_callback_info()));
#ifndef QT_NO_DEBUG
		qDebug("qscan (local) started");
#endif
		goto update_info_end;
	} else if (type == device::DevtypeTCP) {
		if ((threadType != threadDevice) && (threadType != threadMedia))
			goto update_info_err;

#ifndef QT_NO_DEBUG
		qDebug("device: TCP");
#endif
		sock = new QTcpSocket(this);
#ifndef QT_NO_DEBUG
		qDebug("socket created");
#endif
		io->setIODevice(sock);

//		QObject::connect(sock, SIGNAL(readyRead()),
//			this, SLOT(qscan_process_info()));

		sock->connectToHost(host, port);
		if (!sock->waitForConnected(5000)) {
#ifndef QT_NO_DEBUG
			qDebug("Unable to connect to host!");
#endif
			goto update_info_err;
		}

		sock->write("set dev=" + path.toLatin1() + "\n");
		switch (threadType) {
			case threadDevice:
				sock->write("dinfo\n");
				sock->write("close\n");
				break;
			case threadMedia:
				sock->write("minfo\n");
				sock->write("close\n");
				break;
			default:
				goto update_info_err;
		}
		QObject::connect(sock, SIGNAL(disconnected()),
				this, SLOT(qscan_callback_info()));
#ifndef QT_NO_DEBUG
		qDebug("qscan (TCP) started");
#endif
		goto update_info_end;
	}

update_info_err:
	qscan_callback_info();
#ifndef QT_NO_DEBUG
	qDebug() << "END: device::start_update_info(" << threadType << ")";
#endif
	return false;
update_info_end:
#ifndef QT_NO_DEBUG
	qDebug() << "END: device::start_update_info(" << threadType << ")";
#endif
	return true;
}

void device::qscan_callback_info()
{
	int xcode = 0;
#ifndef QT_NO_DEBUG
	qDebug() << "STA: device::qscan_callback_info() " << this;
#endif
	io_mutex->lock();

	QObject::disconnect(io, SIGNAL(readyReadLine()),
		this, SLOT(qscan_process_info()));

	if (type == DevtypeLocal) {
//		QObject::disconnect(proc, SIGNAL(finished(int, QProcess::ExitStatus)),
//				this, SLOT(qscan_callback_info()));
		xcode = proc->exitCode();
#ifndef QT_NO_DEBUG
		qDebug() << "qscan (local) finished: " << xcode;
#endif

		disconnect(proc);
		io->setIODevice(NULL);
	} else if (type == device::DevtypeTCP) {
//		QObject::disconnect(sock, SIGNAL(disconnected()),
//				this, SLOT(qscan_callback_info()));
		sock->disconnectFromHost();
#ifndef QT_NO_DEBUG
		qDebug("qscan (TCP) finished");
#endif
		disconnect(sock);
		io->setIODevice(NULL);
	}
	io_mutex->unlock();
	switch (threadType) {
		case threadDevice:
			emit doneDInfo(xcode);
			break;
		case threadMedia:
			emit doneMInfo(xcode);
			break;
		case threadGetFeatures:
			emit doneGetFeatures(xcode);
			break;
		case threadGetASDB:
			emit doneGetASDB(xcode);
			break;
		default:
			break;
	}

	threadType = threadNone;
	running=0;
	mutex->unlock();
#ifndef QT_NO_DEBUG
	qDebug("END: device::qscan_callback_info()");
#endif
}

bool device::update_device_info()
{
	if (running) return false;
	plugin_names.clear();
	plugin_infos.clear();
	threadType = threadDevice;
	return start();
}

bool device::update_media_info()
{
	if (running) return false;
	threadType = threadMedia;
	return start();
}

bool device::getFeatures()
{
	if (running) return false;
	threadType = threadGetFeatures;
	return start();
}

bool device::getASDB()
{
	if (running) return false;
	threadType = threadGetASDB;
	asdb.clear();
	return start();
}

bool device::startMqck()
{
	if (running) return false;
	threadType = threadMQCK;
	return start();
}

bool device::startAScre()
{
	if (running) return false;
	threadType = threadAScre;
	return start();
}

bool device::startDestruct()
{
	if (running) return false;
	threadType = threadDestruct;
	return start();
}

bool device::startTattoo()
{
	if (running) return false;
	threadType = threadTattoo;
	return start();
}

bool device::setFeature(int f, bool en)
{
	int r=1;
	qDebug() << "device::setFeature: " << f << en;
	if (!mutex->tryLock()) {
#ifndef QT_NO_DEBUG
		qDebug() << "Device busy: " << id;
#endif
		return false;
	}

	if (type == DevtypeLocal) {
		QStringList cdvdopts;
	
		cdvdopts << "-d" << path;
		switch(f) {
			case FEATURE_LOEJ:
				if (en) {
					cdvdopts << "--load";
				} else {
					cdvdopts << "--eject";
				}
				break;
			case FEATURE_LOEJ_TOGGLE:
				cdvdopts << "--loej";
				break;
			case FEATURE_LOCK:
				if (en) {
					cdvdopts << "--lock";
				} else {
					cdvdopts << "--unlock";
				}
				break;
			case FEATURE_LOCK_TOGGLE:
				cdvdopts << "--lockt";
				break;
			case FEATURE_POWEREC:
				cdvdopts << "--powerec";
				cdvdopts << (en ? "on" : "off");
				break;
			case FEATURE_HIDECDR:
				cdvdopts << "--hcdr";
				cdvdopts << (en ? "on" : "off");
				break;
			case FEATURE_SINGLESESSION:
				cdvdopts << "--sss";
				cdvdopts << (en ? "on" : "off");
				break;
			case FEATURE_SPEEDREAD:
				cdvdopts << "--spdread";
				cdvdopts << (en ? "on" : "off");
				break;
			case FEATURE_BITSETR:
				cdvdopts << "--bitset+r";
				cdvdopts << (en ? "on" : "off");
				break;
			case FEATURE_BITSETRDL:
				cdvdopts << "--bitset+rdl";
				cdvdopts << (en ? "on" : "off");
				break;
			case FEATURE_SIMULPLUS:
				cdvdopts << "--dvd+testwrite";
				cdvdopts << (en ? "on" : "off");
				break;
			default:
				mutex->unlock();
				return false;
		}
		qDebug() << cdvdopts.join(" ");
	//	for (int i=0; i<cdvdopts.size();i++) qDebug() << cdvdopts[i];
		r = QProcess::execute("cdvdcontrol", cdvdopts);
	} else if (type == device::DevtypeTCP) {
		switch(f) {
			default:
				mutex->unlock();
				return false;
		}
	}
	mutex->unlock();
	return !!r;
}

bool device::setComplexFeature(int f, DevFeatures* data)
{
	int r=1;
	qDebug() << "device::setComplexFeature: " << f;
	if (!data) {
		qDebug() << "Data pointer is NULL";
		return 1;
	}
	if (!mutex->tryLock()) {
		qDebug() << "Device busy: " << id;
		return 1;
	}

	if (type == DevtypeLocal) {
		QStringList cdvdopts;

		cdvdopts << "-d" << path;
		switch(f) {
			case FEATURE_GIGAREC:
				if (data->enabled & FEATURE_GIGAREC) {
					cdvdopts << "--gigarec" << QString("%1").arg(data->grec, 3, 'f', 1);
				} else {
					cdvdopts << "--gigarec" << "off";
				}
				break;
			case FEATURE_VARIREC_CD:
				if (data->enabled & FEATURE_VARIREC_CD) {
					cdvdopts << "--varirec-cd" << QString::number(data->vrec_cd_pwr);
					if (data->supported & FEATURE_VARIREC_CDEXT) {
						cdvdopts << "--varirec-cd-strategy" << QString::number(data->vrec_cd_str-1);
					}
				} else {
					cdvdopts << "--varirec-cd" << "off";
				}
				break;
			case FEATURE_VARIREC_DVD:
				if (data->enabled & FEATURE_VARIREC_DVD) {
					cdvdopts << "--varirec-dvd" << QString::number(data->vrec_dvd_pwr);
					cdvdopts << "--varirec-dvd-strategy" << QString::number(data->vrec_dvd_str-1);
				} else {
					cdvdopts << "--varirec-dvd" << "off";
				}
				break;
			case FEATURE_SECUREC:
				if (data->enabled & FEATURE_SECUREC) {
					cdvdopts << "--securec" << data->sr_pass;
				} else {
					cdvdopts << "--nosecurec";
				}
				break;
			case FEATURE_SILENT:
				if (data->enabled & FEATURE_SILENT) {
					cdvdopts << "--silent" << "on";
					cdvdopts << "--sm-cd-rd" << QString::number(data->sm_cd_rd);
					cdvdopts << "--sm-cd-wr" << QString::number(data->sm_cd_wr);
					if (cap_rd & DEVICE_DVD) {
						cdvdopts << "--sm-dvd-rd" << QString::number(data->sm_dvd_rd);
//						cdvdopts << "--sm-dvd-wr" << QString::number(data->sm_dvd_wr);
					}
					cdvdopts << "--sm-access" << (data->sm_access ? "fast" : "slow");
					cdvdopts << "--sm-load" << QString::number(data->sm_trayl);
					cdvdopts << "--sm-eject" << QString::number(data->sm_traye);
				} else {
					cdvdopts << "--silent" << "off";
				}
				if (data->sm_nosave) cdvdopts << "--sm-nosave";
				break;
			case FEATURE_AS:
				switch (data->as_action) {
					case AS_ACTION_MODE:
						cdvdopts << "--as-mode";
						switch (data->as_mode) {
							case AS_MODE_OFF:
								cdvdopts << "off";
								break;
							case AS_MODE_AUTO:
								cdvdopts << "auto";
								break;
							case AS_MODE_ON:
								cdvdopts << "on";
								break;
							case AS_MODE_FORCED:
								cdvdopts << "forced";
								break;
							default:
								mutex->unlock();
								return 1;
						}
						break;
					case AS_ACTION_ACT:
						cdvdopts << "--as-on" << QString::number(data->as_idx);
						break;
					case AS_ACTION_DEACT:
						cdvdopts << "--as-off" << QString::number(data->as_idx);
						break;
					case AS_ACTION_DEL:
						cdvdopts << "--as-del" << QString::number(data->as_idx);
						break;
					case AS_ACTION_CLEAR:
						cdvdopts << "--as-clear";
						break;
					default:
						mutex->unlock();
						return 1;
				}
				break;
			case FEATURE_PIOQUIET:
				cdvdopts << "--pio-quiet";
				switch(data->pioq_quiet) {
					case PIOQ_QUIET:
						cdvdopts << "quiet";
						break;
					case PIOQ_STD:
						cdvdopts << "std";
						break;
					case PIOQ_PERF:
						cdvdopts << "perf";
						break;
					default:
						mutex->unlock();
						return 1;
				}
				cdvdopts << "--pio-limit";
				cdvdopts << ((features.enabled & FEATURE_PIOLIMIT) ? "on" : "off");	
				if (data->pioq_nosave)
					cdvdopts << "--pio-nosave";
				break;
			default:
				mutex->unlock();
				return 1;
		}
		qDebug() << cdvdopts.join(" ");
	//	for (int i=0; i<cdvdopts.size();i++) qDebug() << cdvdopts[i];
		r = QProcess::execute("cdvdcontrol", cdvdopts);
	} else if (type == device::DevtypeTCP) {
		switch(f) {
			default:
				mutex->unlock();
				return 1;
		}
	}
	mutex->unlock();
	return !!r;
}

bool device::start_tests() 
{
	if (running) return false;
	threadType = threadTest;
	tests = test_req;
	return start();
}
	
bool device::stop_tests() 
{
	if (!running) return false;
	tests = 0;
	if (type == DevtypeLocal) {
		if (!proc) return false;
		Q_PID pid = proc->pid();
#if defined(__unix) || defined(__unix__)
		kill(pid, SIGINT);
#elif defined(_WIN32)
		TerminateProcess(pid, 0);
#endif
	} else if (type == device::DevtypeTCP) {
		if (!sock) return false;
#ifndef QT_NO_DEBUG
		qDebug("Scan terminate not implemented on network devices");
#endif
		sock->disconnectFromHost();
	}
	return true;
}

bool device::next_test()
{
	ctest=0;
	QString stest;
#ifndef QT_NO_DEBUG
	qDebug("STA: device::next_test()");
#endif

	if (tests & TEST_RT) {
		ctest = TEST_RT;
		stest = "rt";
		testData.clearRT();
		test_spd = tspeeds.rt;
	} else if (tests & TEST_WT) {
		ctest = TEST_WT;
		stest = "wt";
		testData.clearWT();
		test_spd = tspeeds.wt;
	} else if (tests & TEST_ERRC) {
		ctest = TEST_ERRC;
		stest = "errc";
		testData.clearErrc();
		test_spd = tspeeds.errc;
	} else if (tests & TEST_JB) {
		ctest = TEST_JB;
		stest = "jb";
		testData.clearJB();
		test_spd = tspeeds.jb;
	} else if (tests & TEST_FT) {
		ctest = TEST_FT;
		stest = "ft";
		testData.clearTA();
		test_spd = tspeeds.ft;
	} else if (tests & TEST_TA) {
		ctest = TEST_TA;
		stest = "ta";
		testData.clearTA();
//		test_spd = tspeeds.ta;
	}

	tests &= ~ctest;
	if (!ctest) {
		threadType = threadNone;
		running=0;
		mutex->unlock();
#ifndef QT_NO_DEBUG
		qDebug("END: device::next_test(): to tests remaining");
#endif
		emit testsDone();
		return false;
	}

#ifndef QT_NO_DEBUG
		qDebug() << "device::next_test(): starting test " << stest << " at speed " << test_spd;
#endif
/*
	run_test("rt");
	run_test("errc");
	run_test("jb");
	run_test("ft");
	run_test("ta");
*/

	if (type == DevtypeNone) {
		goto next_test_err;
	}
	if (proc) {
		delete proc;
		proc = NULL;
	}
	if (sock) {
		delete sock;
		sock = NULL;
	}

	QObject::connect(io, SIGNAL(readyReadLine()),
		this, SLOT(qscan_process_test()));

	if (type == DevtypeLocal) {
		QStringList qopts;
#ifndef QT_NO_DEBUG
		qDebug("device: LOCAL");
#endif
		proc = new QProcess(this);
#ifndef QT_NO_DEBUG
		qDebug("process created");
#endif
		io->setIODevice(proc);
		proc->setReadChannel(QProcess::StandardOutput);

//		QObject::connect(proc, SIGNAL(readyReadStandardOutput()),
//				this, SLOT(qscan_process_test()));

		qopts << "-d" << path << "-t" << stest << "-s" << QString::number(test_spd);
		if (stest == "wt" && !WT_simul)
			qopts << "-W";
		if (stest != "rt" && stest != "wt" && !plugin.isEmpty()) {
			qopts << "--force-plugin" << plugin;
		}

#if (!defined(QT_NO_DEBUG) && 0)
		for (int i=0;i<qopts.size();i++)
			qDebug("[" + QString::number(i) + "] "+ qopts[i]);
#endif
		proc->start("qscan", qopts);

		if (!proc->waitForStarted(10000)) {
#ifndef QT_NO_DEBUG
			qDebug("Can't run qscan!");
#endif
			goto next_test_err;
		}
		QObject::connect(proc, SIGNAL(finished(int, QProcess::ExitStatus)),
				this, SLOT(qscan_callback_test()));
		gettimeofday(&timeSta, NULL);
#ifndef QT_NO_DEBUG
		qDebug("qscan (local) started");
#endif
		goto next_test_end;
	} else if (type == device::DevtypeTCP) {
#ifndef QT_NO_DEBUG
		qDebug("device: TCP");
#endif
		sock = new QTcpSocket(this);
#ifndef QT_NO_DEBUG
		qDebug("socket created");
#endif
		io->setIODevice(sock);

//		QObject::connect(sock, SIGNAL(readyRead()),
//			this, SLOT(qscan_process_test()));

		sock->connectToHost(host, port);
		if (!sock->waitForConnected(5000)) {
#ifndef QT_NO_DEBUG
			qDebug("Unable to connect to host!");
#endif
			goto next_test_err;
		}
		QObject::connect(sock, SIGNAL(disconnected()),
				this, SLOT(qscan_callback_test()));

		sock->write("set dev=" + path.toLatin1() + "\n");
		sock->write("set test=" + stest.toLatin1() + "\n");
		sock->write("set speed=" + QString::number(test_spd).toLatin1() + "\n");
		if (stest == "wt")
			sock->write("set simul=" + QString::number(WT_simul).toLatin1() + "\n");
		sock->write("run\n");
		sock->write("close\n");
		gettimeofday(&timeSta, NULL);
#ifndef QT_NO_DEBUG
		qDebug("qscan (TCP) started");
#endif
		goto next_test_end;
	}

next_test_err:
	qscan_callback_info();
#ifndef QT_NO_DEBUG
	qDebug("END: device::next_test()");
#endif
	return false;
next_test_end:
#ifndef QT_NO_DEBUG
	qDebug("END: device::next_test()");
#endif
	return true;
}

void device::qscan_process_info()
{
	QString qout;

#ifndef QT_NO_DEBUG
	qDebug("STA: qscan_process_info()");
#endif
	if (!io->IODevice()) {
#ifndef QT_NO_DEBUG
		qDebug("END: qscan_process_info(): QIODevice is NULL");
#endif
		return;
	}
	if (!io_mutex->tryLock()) {
#ifndef QT_NO_DEBUG
		qDebug("END: qscan_process_info(): Can't lock I/O Mutex");
#endif
		return;
	}
	while (io->linesAvailable() ) {
		qout = io->readLine();
//		qout.remove("\n");
		switch (threadType) {
			case threadDevice:
			case threadMedia:
				qscan_process_line(qout);
				break;
			case threadGetFeatures:
				cdvdcontrol_process_line(qout);
				break;
			case threadGetASDB:
				cdvdcontrol_process_asdb(qout);
				break;

			case threadMQCK:
			case threadAScre:
			case threadDestruct:
			case threadTattoo:
			default:
				qDebug() << qout;
				break;
		}
	//	qDebug(qout.remove("\n"));
	} // while (io->bytesAvailable() )
	io_mutex->unlock();
#ifndef QT_NO_DEBUG
	qDebug("END: qscan_process_info()");
#endif
}

void device::qscan_process_line(QString& qout)
{
	QStringList sl;
#ifdef MINFO_TREE
	QTreeWidgetItem *info;
#endif
	QIcon ico_ok(":images/ok.png");
	QIcon ico_x (":images/x.png");
	QIcon ico_rd(":images/disc.png");
	QIcon ico_wr(":images/cdwriter.png");


	if (qout.startsWith("ID:")) {
		// main device params
		qout.remove(0,4);
		sl = qout.split(':');
		if (sl.size() >=2 ) {
			sl[1].remove('\'');
			while (!sl[1].isEmpty() && sl[1][0] == ' ') sl[1].remove(0,1);

			if (sl[0].contains("Device", Qt::CaseInsensitive) && !sl[0].contains("capabilities", Qt::CaseInsensitive)) {
//				l_dev->setText(sl[1]);
#if 0
			} else if (sl[0].contains("Vendor", Qt::CaseInsensitive)) {
				l_vendor->setText(sl[1]);
			} else if (sl[0].contains("Model", Qt::CaseInsensitive)) {
				l_model->setText(sl[1]);
			} else if (sl[0].contains("F/W", Qt::CaseInsensitive)) {
				l_fw->setText(sl[1]);
#endif
			} else if (sl[0].contains("TLA", Qt::CaseInsensitive)) {
				if (sl[1].isEmpty()) {
					tla = "N/A";
				} else {
					tla = sl[1];
				}
			} else if (sl[0].contains("Discs loaded", Qt::CaseInsensitive)) {
				life_dn = sl[1].toInt();
			} else if (sl[0].contains("CD Rd", Qt::CaseInsensitive)) {
				sl.removeAt(0);
				life_cr = sl.join(":");
			} else if (sl[0].contains("CD Wr", Qt::CaseInsensitive)) {
				sl.removeAt(0);
				life_cw = sl.join(":");
			} else if (sl[0].contains("DVD Rd", Qt::CaseInsensitive)) {
				sl.removeAt(0);
				life_dr = sl.join(":");
			} else if (sl[0].contains("DVD Wr", Qt::CaseInsensitive)) {
				sl.removeAt(0);
				life_dw = sl.join(":");
			} else if (sl[0].contains("S/N", Qt::CaseInsensitive)) {
				if (sl[1].isEmpty()) {
					sn = "N/A";
				} else {
					sn = sl[1];
				}
			} else if (sl[0].contains("Buffer", Qt::CaseInsensitive)) {
				buf = sl[1];
			} else if (sl[0].contains("IFace", Qt::CaseInsensitive)) {
				iface = sl[1];
			} else if (sl[0].contains("Loader", Qt::CaseInsensitive)) {
				loader = sl[1];
			} else if (sl[0].contains("RPC Phase", Qt::CaseInsensitive)) {
				rpc_phase = sl[1].toInt();
			} else if (sl[0].contains("Region", Qt::CaseInsensitive)) {
				if (sl[1].contains("not set", Qt::CaseInsensitive))
					rpc_reg = -2;
				else
					rpc_reg = sl[1].toInt();
			} else if (sl[0].contains("Changes left", Qt::CaseInsensitive)) {
				rpc_ch = sl[1].toInt();
			} else if (sl[0].contains("Resets left", Qt::CaseInsensitive)) {
				rpc_rst = sl[1].toInt();
			} else if (sl[0].contains("Device Generic capabilities", Qt::CaseInsensitive)) {
				cap =    sl[1].toULongLong(0,16);
			} else if (sl[0].contains("Device Read capabilities", Qt::CaseInsensitive)) {
				cap_rd = sl[1].toULongLong(0,16);
			} else if (sl[0].contains("Device Write capabilities", Qt::CaseInsensitive)) {
				cap_wr = sl[1].toULongLong(0,16);
			}
		}
	} else if (qout.startsWith("CD:")) {
		// generic device capabilities
		/*
		qout.remove(0,4);
	proc->setReadChannel(QProcess::StandardOutput);
		sl = qout.split(':');
		if (sl.size() >=2 ) {

		}
		*/
	} else if (qout.startsWith("CM:")) {
		// media R/W capabilities
		/*
		bool rd,wr;
		qout.remove(0,4);
		sl = qout.split(':');
			if (sl.size() >=2 ) {
			sl[1].remove('\'');
			rd = sl[1].contains('R', Qt::CaseInsensitive);
			wr = sl[1].contains('W', Qt::CaseInsensitive);
#ifndef QT_NO_DEBUG
//				qDebug("|" + sl[0] + "|" + sl[1]);
#endif
		}
			*/
	} else if (qout.startsWith("IM:")) {
		// device inquiry string
		qout.remove(0,4);
#ifndef QT_NO_DEBUG
		qDebug() << qout;
#endif

		sl = qout.split(':');
		if (sl.size() >=2 ) {
			sl[1].remove('\'');
			while (!sl[1].isEmpty() && sl[1][0] == ' ') sl[1].remove(0,1);
#ifndef QT_NO_DEBUG
//			qDebug("|" + sl[0] + "|" + sl[1] + "|");
#endif
			if (sl[0].contains("Media type", Qt::CaseInsensitive)) {
				if (sl[1].contains("No Media", Qt::CaseInsensitive)) {
					media.type = "-";
					media.spd1X = 1;
				} else {
					media.type = sl[1];
					if (media.type.startsWith("CD")) {
						media.spd1X = 150;
					} else if (media.type.startsWith("DDCD")) {
						media.spd1X = 150;
					} else if (media.type.startsWith("DVD")) {
						media.spd1X = 1385;
					} else if (media.type.startsWith("BD")) {
						media.spd1X = 4495;
					}
				}
			} else if (sl[0].contains("Disc Category", Qt::CaseInsensitive)) {
				media.category = sl[1];
			} else if (sl[0].contains("ERRC speeds", Qt::CaseInsensitive)) {
				media.tspeeds_errc = sl[1].split(" ", QString::SkipEmptyParts);
			} else if (sl[0].contains("JB speeds", Qt::CaseInsensitive)) {
				media.tspeeds_jb = sl[1].split(" ", QString::SkipEmptyParts);
			} else if (sl[0].contains("Layers", Qt::CaseInsensitive)) {
				media.layers = sl[1];
				media.ilayers = sl[1].toInt();
				if (media.ilayers <= 0)
					media.ilayers = 1; 
			} else if (sl[0].contains("Protection", Qt::CaseInsensitive)) {
				media.prot = sl[1];
			} else if (sl[0].contains("Regions", Qt::CaseInsensitive)) {
				media.regions = sl[1];
			} else if (sl[0].contains("Erasable", Qt::CaseInsensitive)) {
				media.erasable = sl[1];
			} else if (sl[0].contains("Disc state", Qt::CaseInsensitive)) {
				media.dstate = sl[1];
			} else if (sl[0].contains("Session state", Qt::CaseInsensitive)) {
				media.sstate = sl[1];
			} else if (sl[0].contains("Read capacity", Qt::CaseInsensitive)) {
				sl.removeFirst();
				QStringList sl2 = sl.join(":").split("/");
				if (sl2.size()>=3) {
					media.creads = sl2[0].remove("sectors").toInt();
					media.creadm = sl2[1].remove("MB").toInt();
					media.creadmsf = sl2[2];
				}
			} else if (sl[0].contains("Free capacity", Qt::CaseInsensitive)) {
				sl.removeFirst();
				QStringList sl2 = sl.join(":").split("/");
				if (sl2.size()>=3) {
					media.cfrees = sl2[0].remove("sectors").toInt();
					media.cfreem = sl2[1].remove("MB").toInt();
					media.cfreemsf = sl2[2];
				}
			} else if (sl[0].contains("Total capacity", Qt::CaseInsensitive)) {
				sl.removeFirst();
				QStringList sl2 = sl.join(":").split("/");
				if (sl2.size()>=3) {
					media.ctots = sl2[0].remove("sectors").toInt();
					media.ctotm = sl2[1].remove("MB").toInt();
					media.ctotmsf = sl2[2];
				}
			} else if (sl[0].contains("Media ID", Qt::CaseInsensitive)) {
				sl.removeFirst();
				media.mid = sl.join(":");
				if (media.type.startsWith("DVD")) media.mid.remove(' ');
			} else if (sl[0].contains("Written on", Qt::CaseInsensitive)) {
				media.writer = sl[1];
#ifdef SHOW_SPEEDS
			} else if (sl[0].contains("RD speed max", Qt::CaseInsensitive)) {
				l_rd_max->setText(sl[1]);
			} else if (sl[0].contains("RD speed #", Qt::CaseInsensitive)) {
				c_rd_lst->addItem(sl[1]);		
			} else if (sl[0].contains("D WR speed max", Qt::CaseInsensitive)) {
				l_wr_max->setText(sl[1]);
			} else if (sl[0].contains("D WR speed #", Qt::CaseInsensitive)) {
				c_wr_lst->addItem(sl[1]);
#endif
			} else if (sl[0].contains("Available quality tests", Qt::CaseInsensitive)) {
				if (sl.size()>=2) {
					QStringList slt = sl[1].split(' ', QString::SkipEmptyParts);
					test_cap = 0;
					for (int ii=0; ii<slt.size(); ii++) {
						test_cap |= (slt[ii] == "errc") ? TEST_ERRC : 0;
						test_cap |= (slt[ii] == "jb")   ? TEST_JB : 0;
						test_cap |= (slt[ii] == "ft")   ? TEST_FT : 0;
						test_cap |= (slt[ii] == "ta")   ? TEST_TA : 0;
					}
				}
			}
		}
#ifdef MINFO_TREE
		info = new QTreeWidgetItem(QStringList(qout));
		info_media.append(info);
#endif
	} else if (qout.startsWith("SM:")) {
		qout.remove(0,4);
#ifndef QT_NO_DEBUG
		qDebug() << qout;
#endif
		sl = qout.split(':');
		if (sl.size() >=2 ) {
			sl[1].remove('\'');
			if (sl[0].contains("RD speed #", Qt::CaseInsensitive)) {
				sl[1].remove(QRegExp("\\([a-z,A-Z,0-9, /]*\\)"));
				sl[1].remove(' ');
				media.rspeeds.append(sl[1]);
			} else if (sl[0].contains("D WR speed #", Qt::CaseInsensitive)) {
				sl[1].remove(QRegExp("\\([a-z,A-Z,0-9, /]*\\)"));
				sl[1].remove(' ');
				media.wspeedsd.prepend(sl[1]);
			} else if (sl[0].contains("M WR speed #", Qt::CaseInsensitive)) {
				sl[1].remove(' ');
				media.wspeedsm.append(sl[1]);
			}

#ifdef MINFO_TREE
			info = new QTreeWidgetItem(QStringList(qout));
			info_media.append(info);
#endif
		}
	} else if (qout.startsWith("Found plugin:")) {
		int spidx;
		QString pn,pi;
		qout.remove(0,13);
		qout = qout.simplified();
		spidx = qout.indexOf(" ");
		pn = qout.mid(0,spidx);
		pi = qout.mid(spidx+1);
		pi.remove("("); pi.remove(")");

		if (!preservePluginsList) {
#ifndef QT_NO_DEBUG
			qDebug() << "Plugin: " << pn << pi;
#endif
			plugin_names.append(pn);
			plugin_infos.append(pi);
		}
	} else if (qout.contains("WARNING!!!  Detected locked PX-755/PX-760")) {
		plextor_lock = 1;
	}
}

void device::cdvdcontrol_process_line(QString& qout)
{
	QStringList sl;
	// vendor-specific device features
//	qDebug() << qout;
	sl = qout.split(':');
	if (sl.size() >=2 ) {
		sl[1].remove(" ");
		if (sl[0].contains("Lock state", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_LOCK;
				features.enabled   |= (sl[1].contains("ON") ? FEATURE_LOCK : 0);
			}
		} else if (sl[0].contains("Hide-CDR", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_HIDECDR;
				features.enabled   |= (sl[1].contains("ON") ? FEATURE_HIDECDR : 0);
			}
		} else if (sl[0].contains("SingleSession", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_SINGLESESSION;
				features.enabled   |= (sl[1].contains("ON") ? FEATURE_SINGLESESSION : 0);
			}
		} else if (sl[0].contains("SpeedRead", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_SPEEDREAD;
				features.enabled   |= (sl[1].contains("ON") ? FEATURE_SPEEDREAD : 0);
			}
		} else if (sl[0].contains("Silent mode", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_SILENT;
				features.enabled   |= (sl[1].contains("ON") ? FEATURE_SILENT : 0);
			}
		} else if (sl[0].contains("SM Read speed", Qt::CaseInsensitive)) {
			features.sm_rd = sl[1].toInt();
		} else if (sl[0].contains("SM Write speed", Qt::CaseInsensitive)) {
			features.sm_wr = sl[1].toInt();
		} else if (sl[0].contains("SM CD Read speed", Qt::CaseInsensitive)) {
			features.sm_cd_rd = sl[1].toInt();
		} else if (sl[0].contains("SM CD Write speed", Qt::CaseInsensitive)) {
			features.sm_cd_wr = sl[1].toInt();
		} else if (sl[0].contains("SM DVD Read speed", Qt::CaseInsensitive)) {
			features.sm_dvd_rd = sl[1].toInt();
//		} else if (sl[0].contains("SM DVD Write speed", Qt::CaseInsensitive)) {
//			features.sm_dvd_wr = sl[1].toInt();
		} else if (sl[0].contains("SM Access time", Qt::CaseInsensitive)) {
			features.sm_access = !!sl[1].contains("FAST");
		} else if (sl[0].contains("SM Eject speed", Qt::CaseInsensitive)) {
			features.sm_traye = sl[1].toInt();
		} else if (sl[0].contains("SM Load speed", Qt::CaseInsensitive)) {
			features.sm_trayl = sl[1].toInt();
		} else if (sl[0].contains("PoweRec", Qt::CaseInsensitive) && !sl[0].contains("Speed", Qt::CaseInsensitive) ) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_POWEREC;
				features.enabled   |= (sl[1].contains("ON") ? FEATURE_POWEREC : 0);
			}
		} else if (sl[0].contains("VariRec CD  power", Qt::CaseInsensitive)) {
			features.vrec_cd_pwr = (char) sl[1].toInt();
		} else if (sl[0].contains("VariRec CD  strategy", Qt::CaseInsensitive)) {
			int offs = sl[1].indexOf('[');
			if (offs>=0) {
				sl[1].remove(0,offs+1);
				sl[1].remove(']');
				features.vrec_cd_str = (char) sl[1].toInt();
			}
		} else if (sl[0].contains("VariRec CD", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_VARIREC_CDBASE;
				if (dev.startsWith("DVDR") || dev.startsWith("CD-R   PREMIUM"))
					features.supported |= FEATURE_VARIREC_CDEXT;
				features.enabled   |= (sl[1].contains("OFF") ? 0 : FEATURE_VARIREC_CD);
			}
		} else if (sl[0].contains("VariRec DVD power", Qt::CaseInsensitive)) {
			features.vrec_dvd_pwr = (char) sl[1].toInt();
		} else if (sl[0].contains("VariRec DVD strategy", Qt::CaseInsensitive)) {
			int offs = sl[1].indexOf('[');
			if (offs>=0) {
				sl[1].remove(0,offs+1);
				sl[1].remove(']');
				features.vrec_dvd_str = (char) sl[1].toInt();
			}
		} else if (sl[0].contains("VariRec DVD", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_VARIREC_DVD;
				features.enabled   |= (sl[1].contains("OFF") ? 0 : FEATURE_VARIREC_DVD);
			}
		} else if (sl[0].contains("GigaRec state", Qt::CaseInsensitive)) {
			features.supported |= FEATURE_GIGAREC;
			if (sl[1].contains("OFF")) {
				features.enabled &= ~FEATURE_GIGAREC;
				features.grec = 1.0;
			} else {
				features.enabled |= FEATURE_GIGAREC;
				features.grec = sl[1].toFloat();
			}
		} else if (sl[0].contains("Disc GigaRec rate", Qt::CaseInsensitive)) {
			if (sl[1].contains("OFF")) {
				media.grec = 1.0;
			} else {
				media.grec = sl[1].toFloat();
			}
		} else if (sl[0].contains("SecuRec", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_SECUREC;
				features.enabled   |= (sl[1].contains("ON") ? FEATURE_SECUREC : 0);
			}
		} else if (sl[0].contains("DVD+R bitset", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_BITSETR;
				features.enabled   |= (sl[1].contains("ON") ? FEATURE_BITSETR : 0);
			}
		} else if (sl[0].contains("DVD+R DL bitset", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_BITSETRDL;
				features.enabled   |= (sl[1].contains("ON") ? FEATURE_BITSETRDL : 0);
			}
		} else if (sl[0].contains("DVD+R(W) testwrite", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_SIMULPLUS;
				features.enabled   |= (sl[1].contains("ON") ? FEATURE_SIMULPLUS : 0);
			}
		} else if (sl[0].contains("AutoStrategy mode", Qt::CaseInsensitive)) {
			features.supported |= FEATURE_AS_BASE;
			if (dev.startsWith("DVDR   PX-755") || dev.startsWith("DVDR   PX-760"))
				features.supported |= FEATURE_AS_EXT;
			if (sl[1].contains("OFF", Qt::CaseInsensitive)) {
				features.as_mode = AS_MODE_OFF;
			} else if (sl[1].contains("AUTO", Qt::CaseInsensitive)) {
				features.as_mode = AS_MODE_AUTO;
			} else if (sl[1].contains("ON", Qt::CaseInsensitive)) {
				features.as_mode = AS_MODE_ON;
			} else if (sl[1].contains("FORCED", Qt::CaseInsensitive)) {
				features.as_mode = AS_MODE_FORCED;
			}
		} else if (sl[0].contains("Pioneer QuietMode", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_PIOQUIET;
				features.supported |= FEATURE_PIOLIMIT;
			}
		} else if (sl[0].contains("PlexEraser", Qt::CaseInsensitive)) {
			if (sl[1] != "-") {
				features.supported |= FEATURE_DESTRUCT;
			}
		} else if (sl[0].contains("Yamaha DiscT@2", Qt::CaseInsensitive)) {
			sl[1].remove("inner");
			sl[1].replace("outer", " ");
			sl[1].replace("image", " ");
			QStringList sl2 = sl[1].split(" ", QString::SkipEmptyParts);
			qDebug () << "*** DISC T@2 ***\nsl[1]: " << sl[1] << "\nsl2 size: " << sl2.size();
			if (sl2.size() >= 3) {
				features.supported |= FEATURE_F1TATTOO;
				features.tattoo_inner = sl2[0].toInt();
				features.tattoo_outer = sl2[1].toInt();
				qDebug() << "Yamaha DiscT@2 radius:  inner " <<  features.tattoo_inner << ", outer " << features.tattoo_outer;
			}
		}
	}
}


void device::cdvdcontrol_process_asdb(QString& qout)
{
	ASDB_item it;
	QStringList sl;
/*
	it.present=1;
	it.active = 1;
	it.type   = "DVD+R";
	it.speed  = QString::number(8);
	it.mid    = "TYG02";
	it.writes = QString::number(25);
	asdb.append(it);
*/
	qDebug("process_asdb");
	if (!qout.startsWith("S#")) return;
	qout.remove(0,2);
	qout.remove(" ");

	sl = qout.split('|');
	if (sl.size() <6) return;

	it.present=1;
	it.active = sl[1] == "*";
//	it.type   = sl[2].remove(QRegExp("[]"));
	it.type   = sl[2].remove("[A1]").remove("[25]");
	it.speed  = sl[3];
	it.mid    = sl[4];
	it.writes = sl[5];

	asdb.append(it);
}

void device::qscan_callback_test()
{
	int xcode = 0;
#ifndef QT_NO_DEBUG
	qDebug("STA: qscan_callback_test()");
#endif
	timeval timeEnd;
	int time;

	qscan_process_test();
	io_mutex->lock();

	QObject::disconnect(io, SIGNAL(readyReadLine()),
		this, SLOT(qscan_process_test()));

	if (type == DevtypeLocal) {
//		QObject::disconnect(proc, SIGNAL(finished(int, QProcess::ExitStatus)),
//				this, SLOT(qscan_callback_info()));
		xcode = proc->exitCode();
#ifndef QT_NO_DEBUG
		qDebug() << "qscan (local) finished" << xcode;
#endif
		disconnect(proc);
		io->setIODevice(NULL);
	} else if (type == device::DevtypeTCP) {
//		QObject::disconnect(sock, SIGNAL(disconnected()),
//				this, SLOT(qscan_callback_info()));
		sock->disconnectFromHost();
#ifndef QT_NO_DEBUG
		qDebug("qscan (TCP) finished");
#endif
		disconnect(sock);
		io->setIODevice(NULL);
	}
	io_mutex->unlock();

	gettimeofday(&timeEnd, NULL);
	time = (int) ((timeEnd.tv_sec - timeSta.tv_sec) + (timeEnd.tv_usec - timeSta.tv_usec)/1000000.0);

	switch(ctest) {
		case TEST_RT:
			testData.rt_time = time;
			emit block_RT();
			break;
		case TEST_WT:
			testData.wt_time = time;
			emit block_WT();
			break;
		case TEST_ERRC:
			testData.errc_time = time;
			emit block_ERRC();
			break;
		case TEST_JB:
			testData.jb_time = time;
			emit block_JB();
			break;
		case TEST_FT:
			testData.ft_time = time;
			emit block_FT();
			break;
		case TEST_TA:
			testData.ta_time = time;
			emit block_TA();
			break;
		default:
			break;
	}

	if (xcode) {
		emit testsError();
	}

	next_test();

#ifndef QT_NO_DEBUG
	qDebug("END: qscan_callback_test()");
#endif
}

void device::qscan_process_test()
{
	timeval timeEnd;
	float time;
	QString qout;
	QStringList sl;
#ifndef QT_NO_DEBUG
//	qDebug("STA: qscan_process_test()");
#endif
	if (!io->IODevice()) {
#ifndef QT_NO_DEBUG
		qDebug("END: qscan_process_test(): QIODevice is NULL");
#endif
		return;
	}
	if (!io_mutex->tryLock()) {
#ifndef QT_NO_DEBUG
		qDebug("END: qscan_process_test(): Can't lock I/O Mutex");
#endif
		return;
	}
	while (io->linesAvailable()) {
		qout = io->readLine();
		qout.remove("\n");
		qout.remove("\r");
#ifndef QT_NO_DEBUG
		qDebug() << qout;
#endif
		gettimeofday(&timeEnd, NULL);
		time = (timeEnd.tv_sec - timeSta.tv_sec) + (timeEnd.tv_usec - timeSta.tv_usec)/1000000.0;
		switch(ctest) {
			case TEST_RT:
				if (qout.startsWith("lba") && qout.contains("speed")) {
//					qDebug("RT");
					DI_Transfer di;
					sl = qout.split(" ", QString::SkipEmptyParts);
			//		for (int i=0; i<sl.size(); i++) qDebug(QString::number(i) + "  '" + sl[i] + "'");
					if (sl.size()>=6) {
						di.lba = sl[1].toLongLong();
						di.spdx = sl[3].toDouble();
						di.spdk = sl[5].toInt();
 						testData.rt.append(di);
						testData.rt_time = time;
						emit block_RT();
					}
			//		qDebug(QString("lba: %1, spdx: %2, spdk: %3").arg(di.lba).arg(di.spdx).arg(di.spdk));
				} else if (qout.startsWith("Reading blocks")) {
					gettimeofday(&timeSta, NULL);
				}
				break;
			case TEST_WT:
				if (qout.startsWith("Track ") && qout.contains("MB written")) {
					DI_Transfer di;
					sl = qout.split(" ", QString::SkipEmptyParts);
			//		for (int i=0; i<sl.size(); i++) qDebug(QString::number(i) + "  '" + sl[i] + "'");
					if (sl.size()>=10) {
						di.lba = sl[2].toLongLong() << 9;
						if (sl[3] == "of") {
							if (sl.size() >=12) {
								di.spdx = sl[11].remove("x.").toDouble();
								di.spdk = (uint32_t)(di.spdx * media.spd1X);
								testData.wt.append(di);
								testData.wt_time = time;
								emit block_WT();
							}
						} else {
							if (sl.size() >=10) {
								di.spdx = sl[9].remove("x.").toDouble();
								di.spdk = (uint32_t)(di.spdx * media.spd1X);
								testData.wt.append(di);
								testData.wt_time = time;
								emit block_WT();
							}
						}
//						qDebug( sl[1] + " " + sl[3] + " " + sl[5] );
					}
			//		qDebug(QString("lba: %1, spdx: %2, spdk: %3").arg(di.lba).arg(di.spdx).arg(di.spdk));
				}
				break;
			case TEST_ERRC:
				if (qout.startsWith("cur")) {
					DI_Errc di;
					sl = qout.split(QRegExp("[:\\ |]"), QString::SkipEmptyParts);
					// for (int i=0; i<sl.size(); i++) qDebug(QString::number(i) + "  '" + sl[i] + "'");

					if (media.type.startsWith("CD") && sl.size()>=14) {
						di.cd.lba = sl[1].toInt();
						di.cd.spdx= sl[2].toFloat();

						di.cd.bler = sl[6].toInt();
						di.cd.e11  = sl[7].toInt();
						di.cd.e21  = sl[8].toInt();
						di.cd.e31  = sl[9].toInt();
						di.cd.e12  = sl[10].toInt();
						di.cd.e22  = sl[11].toInt();
						di.cd.e32  = sl[12].toInt();
						di.cd.uncr = sl[13].toInt();

						ErrcADD(&testData.errcTOT, di);
						ErrcMAX(&testData.errcMAX, di);
						CDErrcAVG(&testData.errcAVG, &testData.errcTOT, di.cd.lba/75);
						testData.errc.append(di);
						testData.errc_time = time;
						emit block_ERRC();
					} else if (media.type.startsWith("DVD") && sl.size()>=13) {
						di.dvd.lba  = sl[1].toInt();
						di.dvd.spdx = sl[2].toFloat();

						di.dvd.res  = 0;
						di.dvd.pie  = sl[6].toInt();
						di.dvd.pi8  = sl[7].toInt();
						di.dvd.pif  = sl[8].toInt();
						di.dvd.poe  = sl[9].toInt();
						di.dvd.po8  = sl[10].toInt();
						di.dvd.pof  = sl[11].toInt();
						di.dvd.uncr = sl[12].toInt();

						ErrcADD(&testData.errcTOT, di);
						ErrcMAX(&testData.errcMAX, di);
						DVDErrcAVG(&testData.errcAVG, &testData.errcTOT, di.dvd.lba >> 4);
						testData.errc.append(di);
						testData.errc_time = time;
						emit block_ERRC();
					}
				}
				break;
			case TEST_JB:
				if (qout.startsWith("cur")) {
					DI_JB di;
					sl = qout.split(QRegExp("[:\\ |]"), QString::SkipEmptyParts);
//					for (int i=0; i<sl.size(); i++) qDebug(QString::number(i) + "'" + sl[i] + "'");

					if (sl.size()>=8) {
						di.lba    = sl[1].toInt();
						di.spdx   = sl[2].toFloat();

						di.jitter = sl[6].toFloat();
						di.asymm  = sl[7].toFloat();

						if (!testData.jb.size()) {
							testData.jbMM.jmin = di.jitter;
							testData.jbMM.jmax = di.jitter;
							testData.jbMM.bmin = di.asymm;
							testData.jbMM.bmax = di.asymm;
						} else {
							if (testData.jbMM.jmin > di.jitter) testData.jbMM.jmin = di.jitter;
							if (testData.jbMM.jmax < di.jitter) testData.jbMM.jmax = di.jitter;
							if (testData.jbMM.bmin > di.asymm)  testData.jbMM.bmin = di.asymm;
							if (testData.jbMM.bmax < di.asymm)  testData.jbMM.bmin = di.asymm;
						}

						testData.jb.append(di);
						testData.jb_time = time;
						emit block_JB();
					}
				}
				break;
			case TEST_FT:
				if (qout.startsWith("cur")) {
					DI_FT di;
					sl = qout.split(QRegExp("[:\\ |]"), QString::SkipEmptyParts);
//					for (int i=0; i<sl.size(); i++) qDebug(QString::number(i) + "'" + sl[i] + "'");

					if (sl.size()>=8) {
						di.lba = sl[1].toInt();
						di.spdx= sl[2].toFloat();

						di.fe  = sl[6].toInt();
						di.te  = sl[7].toInt();

						if (testData.ftMAX.fe < di.fe) testData.ftMAX.fe = di.fe;
						if (testData.ftMAX.te < di.te) testData.ftMAX.te = di.te;

						testData.ft.append(di);
						testData.ft_time = time;
						emit block_FT();
					}
				}
				break;
			case TEST_TA:
				if (qout.startsWith("TA")) {
					if (taIdx <0 || taIdx>5) break;
					DI_TA di;
					sl = qout.split(" ", QString::SkipEmptyParts);
					if (sl.size() >=4) {
						di.idx  = sl[1].toInt();
						di.pit  = sl[2].toInt();
						di.land = sl[3].toInt();
					//	if (!di.idx) testData.ta[taIdx].clear();
						testData.ta[taIdx].append(di);
						testData.ta_time = time;
						emit block_TA();
					}
				} else if (qout.startsWith("Running TA on")) {
					int taLayer=0;
					int taZone=0;
					sl = qout.split(" ", QString::SkipEmptyParts);
					if (sl.size() >=6) {
						sl[3].remove("L");
#ifndef QT_NO_DEBUG
						qDebug() << "TA Layer: " << sl[3];
#endif
						taLayer = sl[3].toInt();
						if (sl[4] == "inner") {
							taZone = 0;
						} else if (sl[4] == "middle") {
							taZone = 1;
						} else if (sl[4] == "outer") {
							taZone = 2;
						}
						taIdx = taLayer*3 + taZone;
					}
				}
				//	parsing TA test output...

				break;
			default:
				break;
		}

	}
	io_mutex->unlock();
#ifndef QT_NO_DEBUG
//	qDebug("END: qscan_process_test()");
#endif
}

bool device::save(QIODevice *f)
{
#ifndef QT_NO_DEBUG
	qDebug("device::save()");
#endif
	QXmlStreamWriter xml;

	if (!f->isOpen()) return false; 
	xml.setDevice(f);

	xml.writeStartDocument();
	xml.writeCharacters("\n");
	xml.writeStartElement("qpxdata");
	xml.writeCharacters("\n");
// writing metadata
// device info...
	xml.writeStartElement("device");
	xml.writeCharacters("\n\t");
	xml.writeStartElement("identify");
	switch (type) {
		case DevtypeLocal:
			xml.writeAttribute("type","local");
			break;
		case DevtypeTCP:
			xml.writeAttribute("type","tcp");
			xml.writeAttribute("host", host);
			xml.writeAttribute("port", QString::number(port));
			break;
		default:
			break;
	}
	xml.writeAttribute("path", path);
//	xml.writeAttribute("id",   id);
	xml.writeAttribute("ven",  ven);
	xml.writeAttribute("dev",  dev);
	xml.writeAttribute("sn",   sn);
	xml.writeAttribute("fw",   fw);
	xml.writeAttribute("tla",  tla);
	xml.writeEndElement();

	xml.writeCharacters("\n\t");
	xml.writeStartElement("info");
	xml.writeAttribute("buf",  buf);
	xml.writeAttribute("iface",iface);
	xml.writeAttribute("loader",loader);
	xml.writeAttribute("cap",   QString("%1").arg(cap,0,16));
	xml.writeAttribute("cap_rd",  QString("%1").arg(cap_rd,0,16));
	xml.writeAttribute("cap_wr", QString("%1").arg(cap_wr,0,16));
	xml.writeEndElement();
	xml.writeCharacters("\n");

	if (cap & CAP_DVD_CSS)
	{
		xml.writeCharacters("\t");
		xml.writeStartElement("rpc");
		xml.writeAttribute("phase",  QString::number(rpc_phase));
		xml.writeAttribute("region", QString::number(rpc_reg));
		xml.writeAttribute("changes", QString::number(rpc_ch));
		xml.writeAttribute("resets", QString::number(rpc_rst));
		xml.writeEndElement();
		xml.writeCharacters("\n");
	}

//	if (life_dn) {
		xml.writeCharacters("\t");
		xml.writeStartElement("lifetime");
		xml.writeAttribute("count",  QString::number(life_dn));
		xml.writeAttribute("cd_rd",  life_cr);
		xml.writeAttribute("cd_wr",  life_cw);
		xml.writeAttribute("dvd_rd", life_dr);
		xml.writeAttribute("dvd_wr", life_dw);
		xml.writeEndElement();
		xml.writeCharacters("\n");
//	}

	xml.writeEndElement(); // device
	xml.writeCharacters("\n");
	xml.writeStartElement("media");
// media info
	xml.writeCharacters("\n\t");
	xml.writeStartElement("identify");
	xml.writeAttribute("type",      media.type);
	xml.writeAttribute("category",  media.category);
	xml.writeAttribute("mid",       media.mid);
	xml.writeEndElement();
	
	xml.writeCharacters("\n\t");
	xml.writeStartElement("capacity");
	xml.writeAttribute("read", QString::number(media.creads));
	xml.writeAttribute("free", QString::number(media.cfrees));
	xml.writeAttribute("tot",  QString::number(media.ctots));

	xml.writeAttribute("read_m", QString::number(media.creadm));
	xml.writeAttribute("free_m", QString::number(media.cfreem));
	xml.writeAttribute("tot_m",  QString::number(media.ctotm));

	xml.writeAttribute("read_msf", media.creadmsf);
	xml.writeAttribute("free_msf", media.cfreemsf);
	xml.writeAttribute("tot_msf",  media.ctotmsf);
	xml.writeEndElement(); // capacity

	xml.writeCharacters("\n\t");
	xml.writeStartElement("speeds");
	xml.writeAttribute("rspeeds",   media.rspeeds.join(" "));
	xml.writeAttribute("wspeedsd",  media.wspeedsd.join(" "));
	xml.writeAttribute("wspeedsm",  media.wspeedsm.join(" "));
	xml.writeEndElement(); // misc

	xml.writeCharacters("\n\t");
	xml.writeStartElement("misc");
	xml.writeAttribute("writer",    media.writer);
	xml.writeAttribute("protection",media.prot);
	xml.writeAttribute("regions",   media.regions);
	xml.writeAttribute("gigarec",   QString("%1").arg(media.grec,0,'f',1));

	xml.writeAttribute("spd1X",     QString::number(media.spd1X));
	xml.writeAttribute("layers",    QString::number(media.ilayers));
	xml.writeAttribute("erasable",  media.erasable);

	xml.writeAttribute("dstate", media.dstate);
	xml.writeAttribute("sstate", media.sstate);
	xml.writeEndElement(); // misc
	
	xml.writeCharacters("\n");

	xml.writeEndElement(); // media
	xml.writeCharacters("\n");
// test speeds

	xml.writeStartElement("tests");
	xml.writeCharacters("\n");

	if (testData.rt.size()) {
		xml.writeCharacters("\t");
		int s = testData.rt.size();
		xml.writeStartElement("rt");
		xml.writeAttribute("speed", QString::number(tspeeds.rt));
		xml.writeAttribute("time",  QString("%1").arg(testData.rt_time,0,'f',2));

		for (int i=0; i<s; i++) {
			xml.writeCharacters("\n\t\t");
			xml.writeTextElement("d", QString("%1,%2,%3").arg(testData.rt[i].lba).arg(testData.rt[i].spdx).arg(testData.rt[i].spdk));
		}
		xml.writeCharacters("\n\t");
		xml.writeEndElement();
		xml.writeCharacters("\n");
	}

	if (testData.wt.size()) {
		xml.writeCharacters("\t");
		int s = testData.rt.size();
		xml.writeStartElement("wt");
		xml.writeAttribute("speed", QString::number(tspeeds.wt));
		xml.writeAttribute("time",  QString("%1").arg(testData.wt_time,0,'f',2));

		for (int i=0; i<s; i++) {
			xml.writeCharacters("\n\t\t");
			xml.writeTextElement("d", QString("%1,%2,%3").arg(testData.wt[i].lba).arg(testData.wt[i].spdx).arg(testData.wt[i].spdk));
		}
		xml.writeCharacters("\n\t");
		xml.writeEndElement();
		xml.writeCharacters("\n");
	}

	if (testData.errc.size()) {
		xml.writeCharacters("\t");
		int s = testData.errc.size();
		xml.writeStartElement("errc");
		xml.writeAttribute("speed", QString::number(tspeeds.errc));
		xml.writeAttribute("time",  QString("%1").arg(testData.errc_time,0,'f',2));
		
		for (int i=0; i<s; i++) {
			xml.writeCharacters("\n\t\t");
			xml.writeTextElement("d", QString("%1,%2,").arg(testData.errc[i].raw.lba).arg(testData.errc[i].raw.spdx,0,'f',2) +
				QString("%1,%2,%3,%4,%5,%6,%7,%8")
					.arg(testData.errc[i].raw.err[0])
					.arg(testData.errc[i].raw.err[1])
					.arg(testData.errc[i].raw.err[2])
					.arg(testData.errc[i].raw.err[3])
					.arg(testData.errc[i].raw.err[4])
					.arg(testData.errc[i].raw.err[5])
					.arg(testData.errc[i].raw.err[6])
					.arg(testData.errc[i].raw.err[7])
					);
		}
		xml.writeCharacters("\n\t");
		xml.writeEndElement();
		xml.writeCharacters("\n");
	}

	if (testData.jb.size()) {
		xml.writeCharacters("\t");
		int s = testData.jb.size();
		xml.writeStartElement("jb");
		xml.writeAttribute("speed", QString::number(tspeeds.jb));
		xml.writeAttribute("time",  QString("%1").arg(testData.jb_time,0,'f',2));

		for (int i=0; i<s; i++) {
			xml.writeCharacters("\n\t\t");
			xml.writeTextElement("d", QString("%1,%2,%3,%4")
					.arg(testData.jb[i].lba)
					.arg(testData.jb[i].spdx)
					.arg(testData.jb[i].jitter)
					.arg(testData.jb[i].asymm));
		}
		xml.writeCharacters("\n\t");
		xml.writeEndElement();
		xml.writeCharacters("\n");
	}

	if (testData.ft.size()) {
		xml.writeCharacters("\t");
		int s = testData.ft.size();
		xml.writeStartElement("ft");
		xml.writeAttribute("speed", QString::number(tspeeds.ft));
		xml.writeAttribute("time",  QString("%1").arg(testData.ft_time,0,'f',2));

		for (int i=0; i<s; i++) {
			xml.writeCharacters("\n\t\t");
			xml.writeTextElement("d", QString("%1,%2,%3,%4")
					.arg(testData.ft[i].lba)
					.arg(testData.ft[i].spdx)
					.arg(testData.ft[i].fe)
					.arg(testData.ft[i].te));
		}
		xml.writeCharacters("\n\t");
		xml.writeEndElement();
		xml.writeCharacters("\n");
	}

	if (testData.ta[0].size()) {
		xml.writeCharacters("\t");
		xml.writeStartElement("ta");
		xml.writeAttribute("time",  QString("%1").arg(testData.ta_time,0,'f',2));

		for (int i=0; i<6; i++) {
			xml.writeCharacters("\n\t\t");
			int s = testData.ta[i].size();
			xml.writeStartElement("zone");
			xml.writeAttribute("idx", QString::number(i));

			for(int j=0;j<s; j++) {
				xml.writeCharacters("\n\t\t\t");
				xml.writeTextElement("d", QString("%1,%2,%3")
					.arg(testData.ta[i][j].idx)
					.arg(testData.ta[i][j].pit)
					.arg(testData.ta[i][j].land)
				);
			}
			xml.writeCharacters("\n\t\t");
			xml.writeEndElement();
		}
		xml.writeEndElement();
		xml.writeCharacters("\n");
	}

	xml.writeEndElement(); // tests
	xml.writeCharacters("\n");
	xml.writeEndElement(); // qpxdata
	xml.writeCharacters("\n");
	xml.writeEndDocument();
	return true;
}

bool device::load(QIODevice *f)
{
#ifndef QT_NO_DEBUG
	qDebug("device::load()");
#endif
	QXmlStreamReader xml;
	QXmlStreamAttributes attr;
	QString name;
	QStringList data;
	bool dataFound = 0,
		 isQpxData = 0,
		 isDevice = 0,
		 isMedia = 0,
		 isTests = 0;
	int  test = 0;
	int	 taZone = -1;

	DI_Transfer d_tr;
	DI_Errc		d_errc;
	DI_JB		d_jb;
	DI_FT		d_ft;
	DI_TA		d_ta;

	if (!f->isOpen()) return false; 
	xml.setDevice(f);
/*
	char		signature[4]; memset(signature, 0, 4);
	if (strncmp(signature, QPX_FILE_SIGN, 4) ) {
		qDebug("Signature missmatch!");
		return false;
	}
*/
	clearMinfo();

	while(!xml.atEnd()) {
		switch (xml.readNext()) {
			case QXmlStreamReader::StartElement:
				name = xml.name().toString();
				if (xml.name() == "qpxdata") {
					dataFound = 1;
					isQpxData = 1;
				}
				if (isQpxData) {
					if (test && (xml.name() == "d")) {
						data = xml.readElementText().split(",");
						switch (test) {
							case TEST_RT:
								if (data.size() == 3) {
									d_tr.lba  = data[0].toLongLong();
									d_tr.spdx = data[1].toFloat();
									d_tr.spdk = data[2].toInt();
									testData.rt << d_tr;
								}
								break;
							case TEST_WT:
								if (data.size() == 3) {
									d_tr.lba  = data[0].toLongLong();
									d_tr.spdx = data[1].toFloat();
									d_tr.spdk = data[2].toInt();
									testData.wt << d_tr;
								}
								break;
							case TEST_ERRC:
								if (data.size() == 10) {
									d_errc.raw.lba    = data[0].toLongLong();
									d_errc.raw.spdx   = data[1].toFloat();
									for (int i=0; i<8; i++)
										d_errc.raw.err[i] = data[2+i].toInt();
#if RECALC_ON_LOAD
									ErrcADD(&testData.errcTOT, d_errc);
									ErrcMAX(&testData.errcMAX, d_errc);
									if (media.type.startsWith("CD")) {
										CDErrcAVG(&testData.errcAVG,  &testData.errcTOT, d_errc.cd.lba/75);
									} else {
										DVDErrcAVG(&testData.errcAVG, &testData.errcTOT, d_errc.dvd.lba >> 4);
									}
#endif
									testData.errc << d_errc;
								}
								break;
							case TEST_JB:
								if (data.size() == 4) {
									d_jb.lba    = data[0].toLongLong();
									d_jb.spdx   = data[1].toFloat();
									d_jb.jitter = data[2].toFloat();
									d_jb.asymm  = data[3].toFloat();
#if RECALC_ON_LOAD
									if (!testData.jb.size()) {
										testData.jbMM.jmin = d_jb.jitter;
										testData.jbMM.jmax = d_jb.jitter;
										testData.jbMM.bmin = d_jb.asymm;
										testData.jbMM.bmax = d_jb.asymm;
									} else {
										if (testData.jbMM.jmin > d_jb.jitter) testData.jbMM.jmin = d_jb.jitter;
										if (testData.jbMM.jmax < d_jb.jitter) testData.jbMM.jmax = d_jb.jitter;
										if (testData.jbMM.bmin > d_jb.asymm)  testData.jbMM.bmin = d_jb.asymm;
										if (testData.jbMM.bmax < d_jb.asymm)  testData.jbMM.bmin = d_jb.asymm;
									}
#endif
									testData.jb << d_jb;
								}
								break;
							case TEST_FT:
								if (data.size() == 4) {
									d_ft.lba = data[0].toLongLong();
									d_ft.spdx= data[1].toFloat();
									d_ft.fe  = data[2].toInt();
									d_ft.te  = data[3].toInt();
#if RECALC_ON_LOAD
									if (testData.ftMAX.fe < d_ft.fe) testData.ftMAX.fe = d_ft.fe;
									if (testData.ftMAX.te < d_ft.te) testData.ftMAX.te = d_ft.te;
#endif
									testData.ft << d_ft;
								}
								break;
							case TEST_TA:
								if (taZone>=0 && taZone<6 && data.size() == 3) {
									d_ta.idx  = data[0].toInt();
									d_ta.pit  = data[1].toInt();
									d_ta.land = data[2].toInt();

									testData.ta[taZone] << d_ta;
								}
								break;
						}
					} else if (isDevice) {
					// device information
						attr = xml.attributes();
						if (xml.name() == "identify") {
//							type = attr.value("type").toInt();
							path = attr.value("path").toString();
							host = attr.value("host").toString();
							port = attr.value("port").toString().toInt();
							path = attr.value("path").toString();
//							id   = attr.value("id").toString();
							ven  = attr.value("ven").toString();
							dev  = attr.value("dev").toString();
							sn   = attr.value("sn").toString();
							fw   = attr.value("fw").toString();
							tla  = attr.value("tla").toString();
						} else if (xml.name() == "info") {
							buf    = attr.value("buf").toString();
							iface  = attr.value("iface").toString();
							loader = attr.value("loader").toString();
							cap    = attr.value("cap").toString().toULongLong(NULL, 16);
							cap_rd = attr.value("cap_rd").toString().toULongLong(NULL, 16);
							cap_wr = attr.value("cap_wr").toString().toULongLong(NULL, 16);
						} else if (xml.name() == "rpc") {
							rpc_phase = attr.value("phase").toString().toInt();
							rpc_reg   = attr.value("region").toString().toInt();
							rpc_ch    = attr.value("changes").toString().toInt();
							rpc_rst   = attr.value("resets").toString().toInt();
						} else if (xml.name() == "lifetime") {
							life_dn = attr.value("count").toString().toInt(); 
							life_cr = attr.value("cd_rd").toString();
							life_cw = attr.value("cd_wr").toString();
							life_dr = attr.value("dvd_rd").toString();
							life_dw = attr.value("dvd_wr").toString();
						}
					} else if (isMedia) {
					// media information
						attr = xml.attributes();
						if (xml.name() == "identify") {
							media.type = attr.value("type").toString();
							media.category = attr.value("category").toString();
							media.mid  = attr.value("mid").toString();
						} else if (xml.name() == "capacity") {
							media.creads = attr.value("read").toString().toInt();
							media.cfrees = attr.value("free").toString().toInt();
							media.ctots  = attr.value("tot").toString().toInt();
							media.creadm = attr.value("read_m").toString().toInt();
							media.cfreem = attr.value("free_m").toString().toInt();
							media.ctotm  = attr.value("tot_m").toString().toInt();
							media.creadmsf = attr.value("read_msf").toString();
							media.cfreemsf = attr.value("free_msf").toString();
							media.ctotmsf  = attr.value("tot_msf").toString();
						} else if (xml.name() == "speeds") {
							media.rspeeds  = attr.value("rspeeds" ).toString().split(" ", QString::SkipEmptyParts);
							media.wspeedsd = attr.value("wspeedsd").toString().split(" ", QString::SkipEmptyParts);
							media.wspeedsm = attr.value("wspeedsm").toString().split(" ", QString::SkipEmptyParts);
						} else if (xml.name() == "misc") {
							media.writer  = attr.value("writer").toString();
							media.prot    = attr.value("protection").toString();
							media.regions = attr.value("regions").toString();
							media.grec    = attr.value("gigarec").toString().toDouble();
							media.spd1X   = attr.value("spd1X").toString().toInt();
							media.layers  = attr.value("layers").toString();
							media.erasable= attr.value("erasable").toString();
							media.ilayers = media.layers.toInt();
							media.dstate  = attr.value("dstate").toString();
							media.sstate  = attr.value("sstate").toString();
						}
					} else if (isTests) {
						attr = xml.attributes();
						if (xml.name() == "rt") {
							test = TEST_RT;
							tspeeds.rt = attr.value("speed").toString().toInt();
							testData.rt_time = attr.value("time").toString().toDouble();
						} else if (xml.name() == "wt") {
							test = TEST_WT;
							tspeeds.wt = attr.value("speed").toString().toInt();
							testData.wt_time = attr.value("time").toString().toDouble();
						} else if (xml.name() == "errc") {
							test = TEST_ERRC;
							tspeeds.errc = attr.value("speed").toString().toInt();
							testData.errc_time = attr.value("time").toString().toDouble();
						} else if (xml.name() == "jb") {
							test = TEST_JB;
							tspeeds.jb = attr.value("speed").toString().toInt();
							testData.jb_time = attr.value("time").toString().toDouble();
						} else if (xml.name() == "ft") {
							test = TEST_FT;
							tspeeds.ft = attr.value("speed").toString().toInt();
							testData.ft_time = attr.value("time").toString().toDouble();
						} else if (xml.name() == "ta") {
							test = TEST_TA;
						//	tspeeds.ta = attr.value("speed").toString().toInt();
							testData.ta_time = attr.value("time").toString().toDouble();
						} else if ((test == TEST_TA) && (xml.name() == "zone")) {
							taZone = attr.value("idx").toString().toInt();
						}
					} else if (xml.name() == "device") {
						isDevice = 1;
					} else if (xml.name() == "media") {
						isMedia = 1;
					} else if (xml.name() == "tests") {
						isTests = 1;
					}
				}
				break;
			case QXmlStreamReader::EndElement:
				if (isQpxData && xml.name() == "qpxdata") {
					isQpxData = 0;
				} else if (isDevice && xml.name() == "device") {
					isDevice = 0;
				} else if (isMedia && xml.name() == "media") {
					isMedia = 0;
				} else if (isTests && xml.name() == "tests") {
					isTests = 0;
				} else if (isTests) {
					if (test == TEST_RT && xml.name() == "rt") {
						test = 0;
					} else if (test == TEST_WT && xml.name() == "wt") {
						test = 0;
					} else if (test == TEST_ERRC && xml.name() == "errc") {
						test = 0;
					} else if (test == TEST_JB && xml.name() == "jb") {
						test = 0;
					} else if (test == TEST_FT && xml.name() == "ft") {
						test = 0;
					} else if (test == TEST_TA && xml.name() == "ta") {
						test = 0;
					} else if (test == TEST_TA && xml.name() == "zone") {
						taZone = -1;
					}
				}
				break;
	//		case QXmlStreamReader::Characters:
	//			break;
			default:
				break;
		}
	}

//	while ( readData(s) ) {}
	emit doneDInfo(0);
	emit doneMInfo(0);

	return dataFound;
}

